// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// This Source Code Form is subject to the terms of the X11 (MIT) License.
// If a copy of the X11 License was not distributed with this file, 
// You can obtain one at https://opensource.org/licenses/MIT/.
//
// X11 (MIT) License
// 
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef VKF_CANDIDATE_HPP
#define VKF_CANDIDATE_HPP

#include "Fragment.hpp" // Declaration of class Fragment
#include "Parents.hpp" // Declaration of class Parents
#include "Sample.hpp" // includes boostdll.h, general.h, and Encoded.hpp too

namespace vkf {
    enum class CandidateType : bool {TOP_CANDIDATE, BOTTOM_CANDIDATE}; //C++11 variant

    class AOMatrix; /**< Class AOMatrix declaration needed for mutual definitions of classes AOMatrix and Candidate */

    class Candidate /**< Class Candidate definition */ {
        friend class AOMatrix; /**< Candidate declares class AOMatrix as its friend class */
        // The private members of class AOMatrix can be accessed because Cadidate is declared as a friend class of AOMatrix too
        protected:
        private:
            Fragment m_Attributes; //!< Member variable "m_Attributes"
            Parents m_Objects; //!< Member variable "m_Objects"
        public:
            /** default Candidate constructor implementation is invoked by AOMatrix constructor */
            Candidate() = default; // C++11-style default constructor definition
            /** copy-constructor */
//            Candidate(const Candidate&) = default; // needed to transfer the upper candidate of Markov chain state
            Candidate(const Candidate&) = delete;
            /** move-constructor */
            Candidate(Candidate&&) = default;
            /** copy-operator= */
            Candidate& operator=(const Candidate&) = delete;
            /** move-operator= */
            Candidate& operator=(Candidate&&) = default;
            /** \brief extreme Candidates constructor declaration
             * \param the first parameter fContext of type AOMatrix contains the bottom and the top candidates
             * \param the second parametr bType determines what type extreme candidate to copy into new candidate */
            Candidate(const AOMatrix& context, const CandidateType& bType);
            /** Incomplete Candidate constructor implementation is invoked by Induction::LoadHypotheses */
            Candidate(boost::dynamic_bitset<boost::dynamic_bitset<>::block_type>&& fragment) : m_Attributes(std::move(fragment)), m_Objects() {}
            /**< default destructor */
            ~Candidate() = default;
            /** \brief Access m_Attributes
             * \return the current value of m_Attributes */
//            Fragment& GetAttributes() const {return m_Attributes;}
            Fragment& GetAttributes() {return m_Attributes;}
            /** \brief Access m_Objects
             * \return the current value of m_Objects */
//            Parents& GetObjects() const {return m_Objects;}
            Parents& GetObjects() {return m_Objects;}
            /** \brief Complete finishes Candidate computation
             * \param context contains the formal context with training examples
             * \param previous_move_up
             * \return second parameter */
            bool Complete(const AOMatrix& context, bool previous_move_up);
            void SetParents(Parents&& parents) {
                m_Objects.clear();
                m_Objects.swap(parents);
            }
            void SetFragment(Fragment&& fragment) {
                m_Attributes.clear();
                m_Attributes.swap(fragment);
            }
    };

    class AOMatrix /**< Class AOMatrix definition */ {
        friend class Candidate; /**< AOMatrix declares class Candidate as its friend class */
        // The private members of class Candidate can be accessed because AOMatrix is declared as a friend class of Candidate too
        protected:
        private:
            std::vector<FragmentPtr> m_Rows; //!< Member variable "m_Rows" (matrix as vector of rows)
            std::vector<ParentsPtr> m_Columns; //!< Member variable "m_Columns" (matrix as vector of columns)
            Candidate m_BottomCandidate; //!< m_BottomCandidate corresponds to the smallest candidate (non-const for AOMAtrix initialization)
            Candidate m_TopCandidate; //!< m_TopCandidate corresponds to the largest candidate (non-const for AOMAtrix initialization)
        public:
            /**< default constructor */
            AOMatrix() = default; // C++11-style default constructor
//            AOMatrix() = delete; // C++11-style forbid the default constructor
            /** copy-constructor */
            AOMatrix(const AOMatrix&) = delete;
            /** move-constructor */
            AOMatrix(AOMatrix&&) = default;
            /** copy-operator= */
            AOMatrix& operator=(const AOMatrix&) = delete;
            /** move-operator= */
            AOMatrix& operator=(AOMatrix&&) = default;
            /**< constructor from Sample declaration */
            AOMatrix(const Sample& rhs, bool is_negative);
            /**< destructor declaration */
            ~AOMatrix();

            /** \brief Derive set of attributes corresponding to given set of objects
             * \param objects is the input set of objects to derive corresponding set of attributes
             * \param result is a variable to store the output set of attributes
             * \return result */
            Fragment& Derive(const Parents& objects, Fragment& result) const;
            /** \brief Derive set of objects corresponding to given set of attributes
             * \param attributes is the input set of attributes to derive corresponding set of objects
             * \param result is a variable to store the output set of objects
             * \return result */
            Parents& Derive(const Fragment& attributes, Parents& result) const;

            /** \brief CbODown intersects given Fragment with attributes of given object
             * \param Fragment is a set of attributes to intersect
             * \param object_number is the number of the object whose attributes participate in intersection
             * \return bit-wise intersection AttrSet attributes with attributes of the object with given number */
            Fragment& CbODown(Fragment& attributes, std::size_t object_number) const {
                attributes &= *(m_Rows.at(object_number));
                return(attributes);
            };

            /** \brief CbOUp intersect given ParentsList with all objects possessing the given attribute
             * \param Parents list of objects to intersect
             * \param attribute_number is the number of the attribute that determines objects participate in intersection
             * \return bit-wise intersection ObjSet objects with objects possessing the attribute with given number*/
            Parents& CbOUp(Parents& objects, std::size_t attribute_number) const {
                objects &= *(m_Columns.at(attribute_number));
                return(objects);
            };

            std::size_t GetRowLength() const {return(m_Columns.size());} //!< Get number of attributes
            std::size_t GetColumnLength() const {return(m_Rows.size());} //!< Get number of objects

            /** \brief IsNextMoveUp determines the direction of next move
             * \param index determines an attibute/object to add
             * \return true if the next move is up */
            bool IsNextMoveUp(std::size_t index) const {return(index >= GetColumnLength());} //!< Determine direction of the next move

            FragmentPtr& GetAttributes(std::size_t index) {return(m_Rows[index]);} //!< Get FragmentPtr for example at index

            static std::mt19937_64 mt_random_engine; //!< Static member variable "mt_random_engine"
            static std::uniform_int_distribution<unsigned long long> index_distribution; //!< Static member variable "index_distribution"
    };

    typedef std::unique_ptr<Candidate> CandidatePtr;
} //namespace vkf

#endif // VKF_CANDIDATE_HPP
