// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// This Source Code Form is subject to the terms of the X11 (MIT) License.
// If a copy of the X11 License was not distributed with this file, 
// You can obtain one at https://opensource.org/licenses/MIT/.
//
// X11 (MIT) License
// 
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef VKF_FCA_HPP
#define VKF_FCA_HPP

#define DEFAULT_HOST "localhost"
#define DEFAULT_PASS "toor"
#define DEFAULT_USER "root"

#include <string>
#include <map>
#include <utility>
#include <vector>
#include <cassert>

#include "Part.hpp"

#include <mariadb++/connection.hpp>
#include <mariadb++/statement.hpp>

namespace vkf {
class FCA : public std::vector<Part> {
    public:
        /** default constructor */
        FCA() = delete;
        FCA(const std::string &lattices_table_name,
                    const std::string &encoder_table_name,
                    const std::string &database,
                    const std::string &url = DEFAULT_HOST,
                    const std::string &user = DEFAULT_USER,
                    const std::string &password = DEFAULT_PASS
                    ) {
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(url, user, password);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            // obtain content of table encoder
            mariadb::result_set_ref result = con->query("SELECT * FROM " + encoder_table_name);
//            std::cout << result->row_count() << std::endl;
            // reading from a result set
            std::string part_name; /**< This variable contains the name of an attribute in current row */
            std::string attribute_value; /**< This variable saves the value in current row */
            std::string value_letter; /**< This variable saves the letter in current row */
            std::uint32_t value_bitset;
            std::uint32_t bitset_length;
            while (result->next()) {
                part_name = result->get_string("attribute_name");
//                std::cout << "part_name: " << part_name << std::endl;
                auto part_ref = std::find_if(this->begin(), this->end(),
                                [&part_name](const Part &x)->bool{ return ( x.getAttributeName() == part_name); });
                if(part_ref == this->end()) /* the attribute does not exist yet */ {
//                    std::cout << "container size: " << this->size() << std::endl;
                    this->emplace_back(Part("part_name"));
//                    std::cout << "container size: " << this->size() << std::endl;
//                    attribute_value = result->get_string("attribute_value");
//                    value_letter = result->get_string("value_letter")[0];
                    this->back().emplace_back(Column(result->get_string("attribute_value"),
                                                    result->get_string("value_letter")[0]));
                } else /* if the attribute exists */ {
                    part_ref->emplace_back(Column(result->get_string("attribute_value"),
                                                    result->get_string("value_letter")[0]));
                }
            }
            // obtain content of table orderings
            result = con->query("SELECT * FROM " + lattices_table_name);
            con->~connection();
            acc->~account();
//            std::cout << result->row_count() << std::endl;
            // reading from a result set
            while (result->next()) {
                part_name = result->get_string("attribute_name");
//                std::cout << "part_name: " << part_name << std::endl;
                auto part_ref = std::find_if(this->begin(), this->end(),
                                [&part_name](const Part &part)->bool{ return ( part.getAttributeName() == part_name); });
                if(part_ref == this->end()) /* the attribute does not exist! */ {
                    throw std::runtime_error("Ordering between values of an nonexistent attribute!");
                } else /* if the attribute exists */ {
                    std::string ssource = result->get_string("source_value");
                    std::size_t columns_number = part_ref->size();
                    std::size_t ndx = columns_number; // index of source value in part's column
                    for (ndx = 0; ndx < columns_number; ndx++) {
                        if (part_ref->at(ndx).getValueName() == ssource) break;
                    }
                    if(ndx == columns_number) /* the source value attribute does not exist! */ {
                        throw std::runtime_error("Source value does not exist!");
                    }
                    std::string starget = result->get_string("target_value");
                    auto target_value_ref = std::find_if(part_ref->begin(), part_ref->end(),
                                [&starget](const Column &column)->bool{ return ( column.getValueName() == starget); });
                    if(target_value_ref == part_ref->end()) /* the source value attribute does not exist! */ {
                        throw std::runtime_error("Target value does not exist!");
                    }
                    target_value_ref->set(ndx, true);
                }
            }
            // set up the account
            acc = mariadb::account::create(url, user, password);
            // create connection
            con = mariadb::connection::create(acc);
            // open database
            mariadb::statement_ref stmt = con->create_statement("INSERT INTO " + encoder_table_name + " VALUES (NULL, ?, ?, ?, ?, ?)");
            for (auto part : *this) {
                part.ComputeMaps();
                part_name = part.getAttributeName();
                for(auto column : part) {
                    attribute_value = column.getValueName();
                    value_letter = column.getSingleLetter();
                    boost::dynamic_bitset<> bdbs = part.findEncoding(attribute_value);
                    value_bitset = bdbs.to_ulong();
                    bitset_length = bdbs.size();
                    stmt->set_string(0, part_name);
                    stmt->set_string(1, attribute_value);
                    stmt->set_string(2, value_letter);
                    stmt->set_unsigned32(3, value_bitset);
                    stmt->set_unsigned32(4, bitset_length);
                    stmt->insert();
                }
            }
//            stmt->~statement();
            con->~connection();
            acc->~account();
        }
        /** default destructor */
        ~FCA()=default;

    private:
};
} // namespace vkf
#endif // VKF_FCA_HPP
