// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// This Source Code Form is subject to the terms of the X11 (MIT) License.
// If a copy of the X11 License was not distributed with this file, 
// You can obtain one at https://opensource.org/licenses/MIT/.
//
// X11 (MIT) License
// 
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include <pybind11/pybind11.h>
#include <pybind11/pytypes.h>

#include "Attributes.hpp"
#include "Data.hpp"
#include "Encoder.hpp"
#include "FCA.hpp"
#include "Generator.hpp"
#include "Graphs.hpp"
#include "Induction.hpp"
#include "Join.hpp"
#include "Lattices.hpp"
#include "Sample.hpp"
#include "TestSample.hpp"

namespace py = pybind11;

PYBIND11_MODULE(vkf, m) {
    m.doc() = "VKF solver procedures";
    py::class_<py::list>(m, "ShowResult");
    py::class_<vkf::CAttributes>(m, "CAttributes")
        .def(py::init<std::string &, std::string &>(),
                "Construct tables for attibutes and their verges",
                py::arg("verges_table_name"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &>(),
                "Construct tables for attibutes and their verges",
                py::arg("verges_table_name"), py::arg("database"),
                py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct tables for attibutes and their verges",
                py::arg("verges_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct tables for attibutes and their verges",
                py::arg("verges_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::DAttributes>(m, "DAttributes")
        .def(py::init<std::string &, std::string &>(),
                "Construct tables for attibutes and their values",
                py::arg("encoder_table_name"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &>(),
                "Construct tables for attibutes and their values",
                py::arg("encoder_table_name"), py::arg("database"),
                py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct tables for attibutes and their values",
                py::arg("encoder_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct tables for attibutes and their values",
                py::arg("encoder_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::Lattices>(m, "Lattices")
        .def(py::init<std::string &, std::string &>(),
                "Construct tables for attibutes and their values",
                py::arg("lattices_table_name"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &>(),
                "Construct tables for attibutes and their values",
                py::arg("encoder_table_name"), py::arg("database"),
                py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct tables for attibutes and their values",
                py::arg("lattices_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct tables for attibutes and their values",
                py::arg("lattices_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::CData>(m, "CData")
        .def(py::init<std::string &, std::string &, std::string &>(),
                "Construct a table for continuous attributes of training/test examples",
                py::arg("data_table_name"), py::arg("verges_table_name"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct a table for continuous attributes of training/test examples",
                py::arg("data_table_name"), py::arg("verges_table_name"), py::arg("database"),
                py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct a table for continuous attributes of training/test examples",
                py::arg("data_table_name"), py::arg("verges_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct a table for continuous attributes of training/test examples",
                py::arg("data_table_name"), py::arg("verges_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::DData>(m, "DData")
        .def(py::init<std::string &, std::string &, std::string &>(),
                "Construct a table for discrete attributes of training/test examples",
                py::arg("data_table_name"), py::arg("encoder_table_name"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct a table for discrete attributes of training/test examples",
                py::arg("data_table_name"), py::arg("encoder_table_name"), py::arg("database"),
                py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct a table for discrete attributes of training/test examples",
                py::arg("data_table_name"), py::arg("encoder_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct a table for discrete attributes of training/test examples",
                py::arg("data_table_name"), py::arg("encoder_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::FData>(m, "FData")
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct a table for both types attributes of training/test examples",
                py::arg("data_table_name"), py::arg("encoder_table_name"), py::arg("verges_table_name"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct a table for both types attributes of training/test examples",
                py::arg("data_table_name"), py::arg("encoder_table_name"), py::arg("verges_table_name"),
                py::arg("database"), py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct a table for both types attributes of training/test examples",
                py::arg("data_table_name"), py::arg("encoder_table_name"), py::arg("verges_table_name"),
                py::arg("database"), py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct a table for both types attributes of training/test examples",
                py::arg("data_table_name"), py::arg("encoder_table_name"), py::arg("verges_table_name"),
                py::arg("database"), py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::FCA>(m, "FCA")
        .def(py::init<std::string &, std::string &, std::string &>(),
                "Complete the table encoder from the table lattices",
                py::arg("lattices_table_name"), py::arg("encoder_table_name"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Complete the table encoder from the table lattices",
                py::arg("lattices_table_name"), py::arg("encoder_table_name"), py::arg("database"),
                py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Complete the table encoder from the table lattices",
                py::arg("lattices_table_name"), py::arg("encoder_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Complete the table encoder from the table lattices",
                py::arg("lattices_table_name"), py::arg("encoder_table_name"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::Graphs>(m, "Graphs")
        .def(py::init<std::string &, std::string &>(),
                    "Create visualizations from the table lattices",
                    py::arg("encoder_table_name"), py::arg("lattices_table_name"))
        .def(py::init<std::string &, std::string &, std::string &>(),
                    "Create visualizations from the table lattices",
                    py::arg("encoder_table_name"), py::arg("lattices_table_name"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                    "Create visualizations from the table lattices",
                    py::arg("encoder_table_name"), py::arg("lattices_table_name"), py::arg("database"),
                    py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                    "Create visualizations from the table lattices",
                    py::arg("encoder_table_name"), py::arg("lattices_table_name"), py::arg("database"),
                    py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &, std::string &>(),
                    "Create visualizations from the table lattices",
                    py::arg("encoder_table_name"), py::arg("lattices_table_name"), py::arg("database"),
                    py::arg("url"), py::arg("user"), py::arg("pass"))
        .def("show_lattices", (py::list (vkf::Graphs::*) ()) &vkf::Graphs::ShowPictures,
                py::return_value_policy::move) // result is moved to new owner
    ;
    py::class_<vkf::Encoder>(m, "Encoder")
        .def(py::init<>(), "Construct empty Encoder")
        .def(py::init<std::string &, std::string &>(),
                "Construct non-empty Encoder",
                py::arg("encoder_table"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &>(),
                "Construct non-empty Encoder",
                py::arg("encoder_table"), py::arg("database"),
                py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Encoder",
                py::arg("encoder_table"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Encoder",
                py::arg("encoder_table"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::Join>(m, "Join")
        .def(py::init<std::string &, std::string &>(),
                "Construct non-empty Join",
                py::arg("data_table"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &>(),
                "Construct non-empty Join",
                py::arg("data_table"), py::arg("database"),
                py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Join",
                py::arg("data_table"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Join",
                py::arg("data_table"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
        .def("compute_save",
                (void (vkf::Join::*) (const std::string &, const std::string &)) &vkf::Join::ComputeSave,
                py::arg("regress_table_name"), py::arg("database"))
        .def("compute_save",
                (void (vkf::Join::*) (const std::string &, const std::string &,
                const std::string &)) &vkf::Join::ComputeSave,
                py::arg("regress_table_name"), py::arg("database"), py::arg("db_host"))
        .def("compute_save",
                (void (vkf::Join::*) (const std::string &, const std::string &,
                const std::string &, const std::string &)) &vkf::Join::ComputeSave,
                py::arg("regress_table_name"), py::arg("database"),
                py::arg("db_host"), py::arg("db_user"))
        .def("compute_save",
                (void (vkf::Join::*) (const std::string &, const std::string &, const std::string &,
                const std::string &, const std::string &)) &vkf::Join::ComputeSave,
                py::arg("regress_table_name"), py::arg("database"),
                py::arg("db_host"), py::arg("db_user"), py::arg("db_pswd"))
    ;
    py::class_<vkf::Generator>(m, "Generator")
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Generator", py::arg("complex_table"), py::arg("data_table"),
                py::arg("verges_table"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::uint32_t &>(),
                "Construct non-empty Generator", py::arg("complex_table"), py::arg("data_table"),
                py::arg("verges_table"), py::arg("database"), py::arg("real_verges_number"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::uint32_t &, std::uint32_t &>(),
                "Construct non-empty Generator", py::arg("complex_table"), py::arg("data_table"),
                py::arg("verges_table"), py::arg("database"), py::arg("real_verges_number"), py::arg("phantom_verges_number"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::uint32_t &, std::uint32_t &,
                std::string &>(), "Construct non-empty Generator",
                py::arg("complex_table"), py::arg("data_table"), py::arg("verges_table"), py::arg("database"),
                py::arg("real_verges_number"), py::arg("phantom_verges_number"), py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::uint32_t &, std::uint32_t &,
                std::string &, std::string &>(), "Construct non-empty Generator",
                py::arg("complex_table"), py::arg("data_table"), py::arg("verges_table"), py::arg("database"),
                py::arg("real_verges_number"), py::arg("phantom_verges_number"), py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::uint32_t &, std::uint32_t &,
                std::string &, std::string &, std::string &>(), "Construct non-empty Generator",
                py::arg("complex_table"), py::arg("data_table"), py::arg("verges_table"), py::arg("database"),
                py::arg("real_verges_number"), py::arg("phantom_verges_number"), py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::Qualifier>(m, "Qualifier")
        .def(py::init<std::string &, std::string &>(),
                "Construct non-empty Qualifier",
                py::arg("verges_table"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &>(),
                "Construct non-empty Qualifier",
                py::arg("verges_table"), py::arg("database"),
                py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Qualifier",
                py::arg("verges_table"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Qualifier",
                py::arg("verges_table"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::Beget>(m, "Beget")
        .def(py::init<std::string &, std::string &>(),
                "Construct non-empty Beget",
                py::arg("complex_table"), py::arg("database"))
        .def(py::init<std::string &, std::string &, std::string &>(),
                "Construct non-empty Beget",
                py::arg("complex_table"), py::arg("database"),
                py::arg("url"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Beget",
                py::arg("complex_table"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Beget",
                py::arg("complex_table"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
    ;
    py::class_<vkf::Sample>(m, "Sample")
        .def(py::init<>(), "Construct empty Sample")
        .def(py::init<vkf::Encoder &, std::string &, std::string &>(),
                "Construct non-empty Sample",
                py::arg("encoder").none(true), py::arg("training_table"), py::arg("database"))
        .def(py::init<vkf::Encoder &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Sample",
                py::arg("encoder").none(true), py::arg("training_table"), py::arg("database"),
                py::arg("url"))
        .def(py::init<vkf::Encoder &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Sample",
                py::arg("encoder").none(true), py::arg("training_table"), py::arg("database"),
                py::arg("url"), py::arg("user"))
        .def(py::init<vkf::Encoder &, std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Sample",
                py::arg("encoder").none(true), py::arg("training_table"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"))
        .def(py::init<vkf::Qualifier &, vkf::Beget &, std::string &, std::string &>(), "Construct non-empty Sample",
                py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("training_table"),
                py::arg("database"), py::keep_alive<1, 2>()) // keep vkf::Qualifier object alive
        .def(py::init<vkf::Qualifier &, vkf::Beget &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Sample",
                py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("training_table"),
                py::arg("database"), py::arg("url"),
                py::keep_alive<1, 2>()) // keep vkf::Qualifier object alive
        .def(py::init<vkf::Qualifier &, vkf::Beget &, std::string &, std::string &, std::string &,
                std::string &>(), "Construct non-empty Sample",
                py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("training_table"),
                py::arg("database"), py::arg("url"), py::arg("user"),
                py::keep_alive<1, 2>()) // keep vkf::Qualifier object alive
        .def(py::init<vkf::Qualifier &, vkf::Beget &, std::string &, std::string &, std::string &,
                std::string &, std::string &>(), "Construct non-empty Sample",
                py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("training_table"),
                py::arg("database"), py::arg("url"), py::arg("user"), py::arg("pass"),
                py::keep_alive<1, 2>()) // keep vkf::Qualifier object alive
        .def(py::init<vkf::Encoder &, vkf::Qualifier &, vkf::Beget &, std::string &, std::string &>(),
                "Construct non-empty Sample",
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("generator").none(true),
                py::arg("training_table"), py::arg("database"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Encoder and vkf::Qualifier objects alive
        .def(py::init<vkf::Encoder &, vkf::Qualifier &, vkf::Beget &, std::string &, std::string &, std::string &>(),
                "Construct non-empty Sample",
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("generator").none(true),
                py::arg("training_table"), py::arg("database"), py::arg("url"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Encoder and vkf::Qualifier objects alive
        .def(py::init<vkf::Encoder &, vkf::Qualifier &, vkf::Beget &, std::string &, std::string &, std::string &,
                std::string &>(), "Construct non-empty Sample",
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("generator").none(true),
                py::arg("training_table"), py::arg("database"), py::arg("url"), py::arg("user"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Encoder and vkf::Qualifier objects alive
        .def(py::init<vkf::Encoder &, vkf::Qualifier &, vkf::Beget &, std::string &, std::string &, std::string &,
                std::string &, std::string &>(), "Construct non-empty Sample",
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("generator").none(true),
                py::arg("training_table"), py::arg("database"), py::arg("url"), py::arg("user"), py::arg("pass"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Encoder and vkf::Qualifier objects alive
    ;
    py::class_<vkf::Induction>(m, "Induction")
        .def(py::init<>(), "Construct empty Induction")
//        .def("insert_sample", (bool (vkf::Induction::*) (vkf::Sample &)) &vkf::Induction::InsertSample,
//                py::arg("sample"))
        .def("load_continuous_hypotheses",
                (int (vkf::Induction::*) (const vkf::Qualifier &, const vkf::Beget &,
                const std::string &, const std::string &, const std::string &)) &vkf::Induction::LoadCHypotheses,
                py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("sample_table_name"),
                py::arg("hypotheses_table_name"), py::arg("database"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Qualifier and vkf.Beget objects alive
        .def("load_continuous_hypotheses",
                (int (vkf::Induction::*) (const vkf::Qualifier &, const vkf::Beget &,
                const std::string &, const std::string &, const std::string &, const std::string &)) &vkf::Induction::LoadCHypotheses,
                py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("sample_table_name"),
                py::arg("hypotheses_table_name"), py::arg("database"), py::arg("db_host"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Qualifier and vkf.Beget objects alive
        .def("load_continuous_hypotheses",
                (int (vkf::Induction::*) (const vkf::Qualifier &, const vkf::Beget &,
                const std::string &, const std::string &, const std::string &, const std::string &,
                const std::string &)) &vkf::Induction::LoadCHypotheses,
                py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("sample_table_name"),
                py::arg("hypotheses_table_name"), py::arg("database"), py::arg("db_host"), py::arg("db_user"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Qualifier and vkf.Beget objects alive
        .def("load_continuous_hypotheses",
                (int (vkf::Induction::*) (const vkf::Qualifier &, const vkf::Beget &,
                const std::string &, const std::string &, const std::string &, const std::string &,
                const std::string &, const std::string &)) &vkf::Induction::LoadCHypotheses,
                py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("sample_table_name"),
                py::arg("hypotheses_table_name"), py::arg("database"), py::arg("db_host"), py::arg("db_user"), py::arg("db_pswd"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Qualifier and vkf.Beget objects alive
        .def("load_discrete_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &,
                const std::string &, const std::string &, const std::string &)) &vkf::Induction::LoadDHypotheses,
                py::arg("encoder").none(true), py::arg("sample_table_name"), py::arg("hypotheses_table_name"),
                py::arg("database"),
                py::keep_alive<1, 2>()) // keep vkf::Encoder object alive
        .def("load_discrete_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &,
                const std::string &, const std::string &, const std::string &, const std::string &)) &vkf::Induction::LoadDHypotheses,
                py::arg("encoder").none(true), py::arg("sample_table_name"), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"),
                py::keep_alive<1, 2>()) // keep vkf::Encoder object alive
        .def("load_discrete_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &,
                const std::string &, const std::string &, const std::string &, const std::string &,
                const std::string &)) &vkf::Induction::LoadDHypotheses,
                py::arg("encoder").none(true), py::arg("sample_table_name"), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"), py::arg("db_user"),
                py::keep_alive<1, 2>()) // keep vkf::Encoder object alive
        .def("load_discrete_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &,
                const std::string &, const std::string &, const std::string &, const std::string &,
                const std::string &, const std::string &)) &vkf::Induction::LoadDHypotheses,
                py::arg("encoder").none(true), py::arg("sample_table_name"), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"), py::arg("db_user"), py::arg("db_pswd"),
                py::keep_alive<1, 2>()) // keep vkf::Encoder object alive
        .def("load_full_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &, const vkf::Qualifier &, const vkf::Beget &,
                const std::string &, const std::string &, const std::string &)) &vkf::Induction::LoadFHypotheses,
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("generator").none(true),
                py::arg("sample_table_name"), py::arg("hypotheses_table_name"), py::arg("database"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>(), py::keep_alive<1, 4>()) // keep vkf::Encoder, vkf::Qualifier and vkf.Beget objects alive
        .def("load_full_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &, const vkf::Qualifier &, const vkf::Beget &,
                const std::string &, const std::string &, const std::string &, const std::string &)) &vkf::Induction::LoadFHypotheses,
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("sample_table_name"),
                py::arg("hypotheses_table_name"), py::arg("database"), py::arg("db_host"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>(), py::keep_alive<1, 4>()) // keep vkf::Encoder, vkf::Qualifier and vkf.Beget objects alive
        .def("load_full_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &, const vkf::Qualifier &, const vkf::Beget &,
                const std::string &, const std::string &, const std::string &, const std::string &,
                const std::string &)) &vkf::Induction::LoadFHypotheses,
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("sample_table_name"),
                py::arg("hypotheses_table_name"), py::arg("database"), py::arg("db_host"), py::arg("db_user"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>(), py::keep_alive<1, 4>()) // keep vkf::Encoder, vkf::Qualifier and vkf.Beget objects alive
        .def("load_full_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &, const vkf::Qualifier &, const vkf::Beget &,
                const std::string &, const std::string &, const std::string &, const std::string &,
                const std::string &, const std::string &)) &vkf::Induction::LoadFHypotheses,
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("generator").none(true), py::arg("sample_table_name"),
                py::arg("hypotheses_table_name"), py::arg("database"), py::arg("db_host"), py::arg("db_user"), py::arg("db_pswd"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>(), py::keep_alive<1, 4>()) // keep vkf::Encoder, vkf::Qualifier and vkf.Beget objects alive
        .def("add_hypotheses", (int (vkf::Induction::*) (int, int)) &vkf::Induction::AddHypotheses,
                py::arg("hypotheses_number"), py::arg("threads_number"))
        .def("save_continuous_hypotheses",
                (int (vkf::Induction::*) (const vkf::Qualifier &,
                const std::string &, const std::string &)) &vkf::Induction::SaveCHypotheses,
                py::arg("encoder").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"))
        .def("save_continuous_hypotheses",
                (int (vkf::Induction::*) (const vkf::Qualifier &,
                const std::string &, const std::string &, const std::string &)) &vkf::Induction::SaveCHypotheses,
                py::arg("encoder").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"))
        .def("save_continuous_hypotheses",
                (int (vkf::Induction::*) (const vkf::Qualifier &,
                const std::string &, const std::string &, const std::string &, const std::string &)) &vkf::Induction::SaveCHypotheses,
                py::arg("encoder").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"), py::arg("db_user"))
        .def("save_continuous_hypotheses",
                (int (vkf::Induction::*) (const vkf::Qualifier &,
                const std::string &, const std::string &, const std::string &, const std::string &,
                const std::string &)) &vkf::Induction::SaveCHypotheses,
                py::arg("encoder").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"), py::arg("db_user"), py::arg("db_pswd"))
        .def("save_discrete_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &,
                const std::string &, const std::string &)) &vkf::Induction::SaveDHypotheses,
                py::arg("encoder").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"))
        .def("save_discrete_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &,
                const std::string &, const std::string &, const std::string &)) &vkf::Induction::SaveDHypotheses,
                py::arg("encoder").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"))
        .def("save_discrete_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &,
                const std::string &, const std::string &, const std::string &, const std::string &)) &vkf::Induction::SaveDHypotheses,
                py::arg("encoder").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"), py::arg("db_user"))
        .def("save_discrete_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &,
                const std::string &, const std::string &, const std::string &, const std::string &,
                const std::string &)) &vkf::Induction::SaveDHypotheses,
                py::arg("encoder").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"), py::arg("db_user"), py::arg("db_pswd"))
        .def("save_full_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &, const vkf::Qualifier &,
                const std::string &, const std::string &)) &vkf::Induction::SaveFHypotheses,
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"))
        .def("save_full_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &, const vkf::Qualifier &,
                const std::string &, const std::string &, const std::string &)) &vkf::Induction::SaveFHypotheses,
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"))
        .def("save_full_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &, const vkf::Qualifier &,
                const std::string &, const std::string &, const std::string &, const std::string &))
                &vkf::Induction::SaveFHypotheses,
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"), py::arg("db_user"))
        .def("save_full_hypotheses",
                (int (vkf::Induction::*) (const vkf::Encoder &, const vkf::Qualifier &,
                const std::string &, const std::string &, const std::string &, const std::string &,
                const std::string &)) &vkf::Induction::SaveFHypotheses,
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("hypotheses_table_name"),
                py::arg("database"), py::arg("db_host"), py::arg("db_user"), py::arg("db_pswd"))
        .def("get_size", (int (vkf::Induction::*) ()) &vkf::Induction::GetSize)
        .def("show_continuous_hypothesis", (py::list (vkf::Induction::*) (const vkf::Qualifier &, int)) &vkf::Induction::ShowCHypothesis,
                py::arg("thresholds").none(true), py::arg("ndx"), py::return_value_policy::move) // result is moved to new owner
        .def("show_discrete_hypothesis", (py::list (vkf::Induction::*) (const vkf::Encoder &, int)) &vkf::Induction::ShowDHypothesis,
                py::arg("encoder").none(true), py::arg("ndx"), py::return_value_policy::move) // result is moved to new owner
        .def("show_full_hypothesis", (py::list (vkf::Induction::*) (const vkf::Encoder &, const vkf::Qualifier &, int)) &vkf::Induction::ShowFHypothesis,
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("ndx"), py::return_value_policy::move) // result is moved to new owner
    ;
    py::class_<vkf::TestSample>(m, "TestSample")
        .def(py::init<vkf::Encoder &, vkf::Induction &, std::string &, std::string &>(),
                "Construct non-empty TestSample",
                py::arg("encoder").none(true), py::arg("causes"), py::arg("tests_table"), py::arg("database"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Encoder and vkf::Induction objects alive
        .def(py::init<vkf::Encoder &, vkf::Induction &, std::string &, std::string &, std::string &>(),
                "Construct non-empty TestSample",
                py::arg("encoder").none(true), py::arg("causes"), py::arg("tests_table"), py::arg("database"),
                py::arg("url"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Encoder and vkf::Induction objects alive
        .def(py::init<vkf::Encoder &, vkf::Induction &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty TestSample",
                py::arg("encoder").none(true), py::arg("causes"), py::arg("tests_table"), py::arg("database"),
                py::arg("url"), py::arg("user"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Encoder and vkf::Induction objects alive
        .def(py::init<vkf::Encoder &, vkf::Induction &, std::string &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty TestSample",
                py::arg("encoder").none(true), py::arg("causes"), py::arg("tests_table"), py::arg("database"),
                py::arg("url"), py::arg("user"), py::arg("pass"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Encoder and vkf::Induction objects alive
        .def(py::init<vkf::Qualifier &, vkf::Induction &, vkf::Beget &, std::string &, std::string &>(), "Construct non-empty TestSample",
                py::arg("thresholds").none(true), py::arg("causes").none(true), py::arg("generator").none(true), py::arg("tests_table"),
                py::arg("database"), py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Qualifier and vkf::Induction objects alive
        .def(py::init<vkf::Qualifier &, vkf::Induction &, vkf::Beget &, std::string &, std::string &, std::string &>(),
                "Construct non-empty TestSample",
                py::arg("thresholds").none(true), py::arg("causes").none(true), py::arg("generator").none(true), py::arg("tests_table"),
                py::arg("database"), py::arg("url"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Qualifier and vkf::Induction objects alive
        .def(py::init<vkf::Qualifier &, vkf::Induction &, vkf::Beget &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty TestSample",
                py::arg("thresholds").none(true), py::arg("causes").none(true), py::arg("generator").none(true), py::arg("tests_table"),
                py::arg("database"), py::arg("url"), py::arg("user"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Qualifier and vkf::Induction objects alive
        .def(py::init<vkf::Qualifier &, vkf::Induction &, vkf::Beget &, std::string &, std::string &, std::string &, std::string &,
                std::string &>(), "Construct non-empty TestSample",
                py::arg("thresholds").none(true), py::arg("causes").none(true), py::arg("generator").none(true), py::arg("tests_table"),
                py::arg("database"), py::arg("url"), py::arg("user"), py::arg("pass"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>()) // keep vkf::Qualifier and vkf::Induction objects alive
        .def(py::init<vkf::Encoder &, vkf::Qualifier &, vkf::Induction &, vkf::Beget &, std::string &, std::string &>(), "Construct non-empty TestSample",
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("causes").none(true), py::arg("generator").none(true),
                py::arg("tests_table"), py::arg("database"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>(), py::keep_alive<1, 4>()) // keep vkf::Encoder, vkf::Qualifier and vkf::Induction objects alive
        .def(py::init<vkf::Encoder &, vkf::Qualifier &, vkf::Induction &, vkf::Beget &, std::string &, std::string &, std::string &>(),
                "Construct non-empty TestSample",
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("causes").none(true), py::arg("generator").none(true),
                py::arg("tests_table"), py::arg("database"), py::arg("url"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>(), py::keep_alive<1, 4>()) // keep vkf::Encoder, vkf::Qualifier and vkf::Induction objects alive
        .def(py::init<vkf::Encoder &, vkf::Qualifier &, vkf::Induction &, vkf::Beget &, std::string &, std::string &, std::string &, std::string &>(),
                "Construct non-empty TestSample",
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("causes").none(true), py::arg("generator").none(true),
                py::arg("tests_table"), py::arg("database"), py::arg("url"), py::arg("user"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>(), py::keep_alive<1, 4>()) // keep vkf::Encoder, vkf::Qualifier and vkf::Induction objects alive
        .def(py::init<vkf::Encoder &, vkf::Qualifier &, vkf::Induction &, vkf::Beget &, std::string &, std::string &, std::string &, std::string &,
                std::string &>(), "Construct non-empty TestSample",
                py::arg("encoder").none(true), py::arg("thresholds").none(true), py::arg("causes").none(true), py::arg("generator").none(true),
                py::arg("tests_table"), py::arg("database"), py::arg("url"), py::arg("user"), py::arg("pass"),
                py::keep_alive<1, 2>(), py::keep_alive<1, 3>(), py::keep_alive<1, 4>()) // keep vkf::Encoder, vkf::Qualifier and vkf::Induction objects alive
        .def("correct_positive_cases", (int (vkf::TestSample::*) ()) &vkf::TestSample::GetPositive)
        .def("correct_negative_cases", (int (vkf::TestSample::*) ()) &vkf::TestSample::GetNegative)
    ;
}
